"""ProductRegistry contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class ProductState(enum.IntEnum):
    """Port of `enum ProductState` on the ProductRegistry contract."""

    NOT_EXIST = 0
    PENDING = 1
    LIVE = 2
    TRADEOUT = 3
    FINAL_SETTLEMENT = 4
    EXPIRED = 5


@dataclass
class ProductMetadata:
    """Port of `struct ProductMetadata` on the IProductRegistry contract."""

    builder: eth_typing.ChecksumAddress
    symbol: str
    description: str


@dataclass
class OracleSpecification:
    """Port of `struct OracleSpecification` on the IProductRegistry contract."""

    oracle_address: eth_typing.ChecksumAddress
    fsv_decimals: int
    fsp_alpha: int
    fsp_beta: int
    fsv_calldata: hexbytes.HexBytes


@dataclass
class Product:
    """Port of `struct Product` on the IProductRegistry contract."""

    metadata: ProductMetadata
    oracle_spec: OracleSpecification
    price_quotation: str
    collateral_asset: eth_typing.ChecksumAddress
    start_time: int
    earliest_fsp_submission_time: int
    unit_value: int
    initial_margin_requirement: int
    maintenance_margin_requirement: int
    offer_price_buffer: int
    auction_bounty: int
    tradeout_interval: int
    tick_size: int
    extended_metadata: str


class ProductRegistry:
    """ProductRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ProductRegistry contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the ProductRegistry contract."""
        return self._contract.events.Initialized

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the ProductRegistry contract."""
        return self._contract.events.OwnershipTransferred

    @property
    def ProductRegistered(self) -> contract.ContractEvent:
        """Binding for `event ProductRegistered` on the ProductRegistry contract."""
        return self._contract.events.ProductRegistered

    @property
    def Upgraded(self) -> contract.ContractEvent:
        """Binding for `event Upgraded` on the ProductRegistry contract."""
        return self._contract.events.Upgraded

    def upgrade_interface_version(
        self,
    ) -> str:
        """Binding for `UPGRADE_INTERFACE_VERSION` on the ProductRegistry contract.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.UPGRADE_INTERFACE_VERSION().call()
        return str(return_value)

    def auction_bounty(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `auctionBounty` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.auctionBounty(
            product_id,
        ).call()
        return int(return_value)

    def clearing(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the ProductRegistry contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call()
        return eth_typing.ChecksumAddress(return_value)

    def collateral_asset(
        self,
        product_id: hexbytes.HexBytes,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateralAsset` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateralAsset(
            product_id,
        ).call()
        return eth_typing.ChecksumAddress(return_value)

    def earliest_fsp_submission_time(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `earliestFSPSubmissionTime` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.earliestFSPSubmissionTime(
            product_id,
        ).call()
        return int(return_value)

    def id(
        self,
        product: Product,
    ) -> hexbytes.HexBytes:
        """Binding for `id` on the ProductRegistry contract.

        Parameters
        ----------
        product : Product

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.id(
            (
                (
                    product.metadata.builder,
                    product.metadata.symbol,
                    product.metadata.description,
                ),
                (
                    product.oracle_spec.oracle_address,
                    product.oracle_spec.fsv_decimals,
                    product.oracle_spec.fsp_alpha,
                    product.oracle_spec.fsp_beta,
                    product.oracle_spec.fsv_calldata,
                ),
                product.price_quotation,
                product.collateral_asset,
                product.start_time,
                product.earliest_fsp_submission_time,
                product.unit_value,
                product.initial_margin_requirement,
                product.maintenance_margin_requirement,
                product.offer_price_buffer,
                product.auction_bounty,
                product.tradeout_interval,
                product.tick_size,
                product.extended_metadata,
            ),
        ).call()
        return hexbytes.HexBytes(return_value)

    def imr(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `imr` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.imr(
            product_id,
        ).call()
        return int(return_value)

    def initialize(
        self,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the ProductRegistry contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize()

    def mmr(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `mmr` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mmr(
            product_id,
        ).call()
        return int(return_value)

    def offer_price_buffer(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `offerPriceBuffer` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.offerPriceBuffer(
            product_id,
        ).call()
        return int(return_value)

    def oracle_specification(
        self,
        product_id: hexbytes.HexBytes,
    ) -> OracleSpecification:
        """Binding for `oracleSpecification` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        OracleSpecification
        """
        return_value = self._contract.functions.oracleSpecification(
            product_id,
        ).call()
        return OracleSpecification(
            eth_typing.ChecksumAddress(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            hexbytes.HexBytes(return_value[4]),
        )

    def owner(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the ProductRegistry contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call()
        return eth_typing.ChecksumAddress(return_value)

    def point_value(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `pointValue` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.pointValue(
            product_id,
        ).call()
        return int(return_value)

    def products(
        self,
        product_id: hexbytes.HexBytes,
    ) -> Product:
        """Binding for `products` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        Product
        """
        return_value = self._contract.functions.products(
            product_id,
        ).call()
        return Product(
            ProductMetadata(
                eth_typing.ChecksumAddress(return_value[0][0]),
                str(return_value[0][1]),
                str(return_value[0][2]),
            ),
            OracleSpecification(
                eth_typing.ChecksumAddress(return_value[1][0]),
                int(return_value[1][1]),
                int(return_value[1][2]),
                int(return_value[1][3]),
                hexbytes.HexBytes(return_value[1][4]),
            ),
            str(return_value[2]),
            eth_typing.ChecksumAddress(return_value[3]),
            int(return_value[4]),
            int(return_value[5]),
            int(return_value[6]),
            int(return_value[7]),
            int(return_value[8]),
            int(return_value[9]),
            int(return_value[10]),
            int(return_value[11]),
            int(return_value[12]),
            str(return_value[13]),
        )

    def proxiable_uuid(
        self,
    ) -> hexbytes.HexBytes:
        """Binding for `proxiableUUID` on the ProductRegistry contract.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.proxiableUUID().call()
        return hexbytes.HexBytes(return_value)

    def register(
        self,
        product: Product,
    ) -> contract.ContractFunction:
        """Binding for `register` on the ProductRegistry contract.

        Parameters
        ----------
        product : Product

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.register(
            (
                (
                    product.metadata.builder,
                    product.metadata.symbol,
                    product.metadata.description,
                ),
                (
                    product.oracle_spec.oracle_address,
                    product.oracle_spec.fsv_decimals,
                    product.oracle_spec.fsp_alpha,
                    product.oracle_spec.fsp_beta,
                    product.oracle_spec.fsv_calldata,
                ),
                product.price_quotation,
                product.collateral_asset,
                product.start_time,
                product.earliest_fsp_submission_time,
                product.unit_value,
                product.initial_margin_requirement,
                product.maintenance_margin_requirement,
                product.offer_price_buffer,
                product.auction_bounty,
                product.tradeout_interval,
                product.tick_size,
                product.extended_metadata,
            ),
        )

    def renounce_ownership(
        self,
    ) -> contract.ContractFunction:
        """Binding for `renounceOwnership` on the ProductRegistry contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceOwnership()

    def set_clearing(
        self,
        clearing_: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `setClearing` on the ProductRegistry contract.

        Parameters
        ----------
        clearing_ : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setClearing(
            clearing_,
        )

    def state(
        self,
        product_id: hexbytes.HexBytes,
    ) -> ProductState:
        """Binding for `state` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        ProductState
        """
        return_value = self._contract.functions.state(
            product_id,
        ).call()
        return ProductState(return_value)

    def tick_size(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `tickSize` on the ProductRegistry contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.tickSize(
            product_id,
        ).call()
        return int(return_value)

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the ProductRegistry contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )

    def upgrade_to_and_call(
        self,
        new_implementation: eth_typing.ChecksumAddress,
        data: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `upgradeToAndCall` on the ProductRegistry contract.

        Parameters
        ----------
        new_implementation : eth_typing.ChecksumAddress
        data : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.upgradeToAndCall(
            new_implementation,
            data,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
        {
            "inputs": [
                {"internalType": "address", "name": "target", "type": "address"}
            ],
            "name": "AddressEmptyCode",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "parameter", "type": "string"}
            ],
            "name": "AlreadyExists",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "implementation", "type": "address"}
            ],
            "name": "ERC1967InvalidImplementation",
            "type": "error",
        },
        {"inputs": [], "name": "ERC1967NonPayable", "type": "error"},
        {"inputs": [], "name": "FailedCall", "type": "error"},
        {
            "inputs": [
                {"internalType": "uint256", "name": "startTime", "type": "uint256"},
                {
                    "internalType": "uint256",
                    "name": "earliestFSPSubmissionTime",
                    "type": "uint256",
                },
            ],
            "name": "InvalidFSPSubmissionTime",
            "type": "error",
        },
        {"inputs": [], "name": "InvalidInitialization", "type": "error"},
        {
            "inputs": [
                {"internalType": "uint16", "name": "imr", "type": "uint16"},
                {"internalType": "uint16", "name": "mmr", "type": "uint16"},
            ],
            "name": "InvalidMarginRequirement",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "startTime", "type": "uint256"},
                {
                    "internalType": "uint256",
                    "name": "blockTimestamp",
                    "type": "uint256",
                },
            ],
            "name": "InvalidStartTime",
            "type": "error",
        },
        {"inputs": [], "name": "NotInitializing", "type": "error"},
        {
            "inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
            "name": "OwnableInvalidOwner",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "OwnableUnauthorizedAccount",
            "type": "error",
        },
        {"inputs": [], "name": "UUPSUnauthorizedCallContext", "type": "error"},
        {
            "inputs": [{"internalType": "bytes32", "name": "slot", "type": "bytes32"}],
            "name": "UUPSUnsupportedProxiableUUID",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "uint64",
                    "name": "version",
                    "type": "uint64",
                }
            ],
            "name": "Initialized",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "previousOwner",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "newOwner",
                    "type": "address",
                },
            ],
            "name": "OwnershipTransferred",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "builder",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "bytes32",
                    "name": "productId",
                    "type": "bytes32",
                },
            ],
            "name": "ProductRegistered",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "implementation",
                    "type": "address",
                }
            ],
            "name": "Upgraded",
            "type": "event",
        },
        {
            "inputs": [],
            "name": "UPGRADE_INTERFACE_VERSION",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "auctionBounty",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "clearing",
            "outputs": [
                {
                    "internalType": "contract IClearingSubset",
                    "name": "",
                    "type": "address",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "collateralAsset",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "earliestFSPSubmissionTime",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "builder",
                                    "type": "address",
                                },
                                {
                                    "internalType": "string",
                                    "name": "symbol",
                                    "type": "string",
                                },
                                {
                                    "internalType": "string",
                                    "name": "description",
                                    "type": "string",
                                },
                            ],
                            "internalType": "struct IProductRegistry.ProductMetadata",
                            "name": "metadata",
                            "type": "tuple",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "oracleAddress",
                                    "type": "address",
                                },
                                {
                                    "internalType": "uint8",
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspAlpha",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspBeta",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IProductRegistry.OracleSpecification",
                            "name": "oracleSpec",
                            "type": "tuple",
                        },
                        {
                            "internalType": "string",
                            "name": "priceQuotation",
                            "type": "string",
                        },
                        {
                            "internalType": "address",
                            "name": "collateralAsset",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "startTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "earliestFSPSubmissionTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "unitValue",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint16",
                            "name": "initialMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint16",
                            "name": "maintenanceMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint64",
                            "name": "offerPriceBuffer",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint64",
                            "name": "auctionBounty",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint32",
                            "name": "tradeoutInterval",
                            "type": "uint32",
                        },
                        {"internalType": "uint8", "name": "tickSize", "type": "uint8"},
                        {
                            "internalType": "string",
                            "name": "extendedMetadata",
                            "type": "string",
                        },
                    ],
                    "internalType": "struct IProductRegistry.Product",
                    "name": "product",
                    "type": "tuple",
                }
            ],
            "name": "id",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "pure",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "imr",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "initialize",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "mmr",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "offerPriceBuffer",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "oracleSpecification",
            "outputs": [
                {
                    "components": [
                        {
                            "internalType": "address",
                            "name": "oracleAddress",
                            "type": "address",
                        },
                        {
                            "internalType": "uint8",
                            "name": "fsvDecimals",
                            "type": "uint8",
                        },
                        {
                            "internalType": "int256",
                            "name": "fspAlpha",
                            "type": "int256",
                        },
                        {"internalType": "int256", "name": "fspBeta", "type": "int256"},
                        {
                            "internalType": "bytes",
                            "name": "fsvCalldata",
                            "type": "bytes",
                        },
                    ],
                    "internalType": "struct IProductRegistry.OracleSpecification",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "owner",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "pointValue",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "products",
            "outputs": [
                {
                    "components": [
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "builder",
                                    "type": "address",
                                },
                                {
                                    "internalType": "string",
                                    "name": "symbol",
                                    "type": "string",
                                },
                                {
                                    "internalType": "string",
                                    "name": "description",
                                    "type": "string",
                                },
                            ],
                            "internalType": "struct IProductRegistry.ProductMetadata",
                            "name": "metadata",
                            "type": "tuple",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "oracleAddress",
                                    "type": "address",
                                },
                                {
                                    "internalType": "uint8",
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspAlpha",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspBeta",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IProductRegistry.OracleSpecification",
                            "name": "oracleSpec",
                            "type": "tuple",
                        },
                        {
                            "internalType": "string",
                            "name": "priceQuotation",
                            "type": "string",
                        },
                        {
                            "internalType": "address",
                            "name": "collateralAsset",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "startTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "earliestFSPSubmissionTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "unitValue",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint16",
                            "name": "initialMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint16",
                            "name": "maintenanceMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint64",
                            "name": "offerPriceBuffer",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint64",
                            "name": "auctionBounty",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint32",
                            "name": "tradeoutInterval",
                            "type": "uint32",
                        },
                        {"internalType": "uint8", "name": "tickSize", "type": "uint8"},
                        {
                            "internalType": "string",
                            "name": "extendedMetadata",
                            "type": "string",
                        },
                    ],
                    "internalType": "struct IProductRegistry.Product",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "proxiableUUID",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "builder",
                                    "type": "address",
                                },
                                {
                                    "internalType": "string",
                                    "name": "symbol",
                                    "type": "string",
                                },
                                {
                                    "internalType": "string",
                                    "name": "description",
                                    "type": "string",
                                },
                            ],
                            "internalType": "struct IProductRegistry.ProductMetadata",
                            "name": "metadata",
                            "type": "tuple",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "oracleAddress",
                                    "type": "address",
                                },
                                {
                                    "internalType": "uint8",
                                    "name": "fsvDecimals",
                                    "type": "uint8",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspAlpha",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "fspBeta",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "fsvCalldata",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IProductRegistry.OracleSpecification",
                            "name": "oracleSpec",
                            "type": "tuple",
                        },
                        {
                            "internalType": "string",
                            "name": "priceQuotation",
                            "type": "string",
                        },
                        {
                            "internalType": "address",
                            "name": "collateralAsset",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "startTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "earliestFSPSubmissionTime",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "unitValue",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint16",
                            "name": "initialMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint16",
                            "name": "maintenanceMarginRequirement",
                            "type": "uint16",
                        },
                        {
                            "internalType": "uint64",
                            "name": "offerPriceBuffer",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint64",
                            "name": "auctionBounty",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint32",
                            "name": "tradeoutInterval",
                            "type": "uint32",
                        },
                        {"internalType": "uint8", "name": "tickSize", "type": "uint8"},
                        {
                            "internalType": "string",
                            "name": "extendedMetadata",
                            "type": "string",
                        },
                    ],
                    "internalType": "struct IProductRegistry.Product",
                    "name": "product",
                    "type": "tuple",
                }
            ],
            "name": "register",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "renounceOwnership",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "contract IClearingSubset",
                    "name": "clearing_",
                    "type": "address",
                }
            ],
            "name": "setClearing",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "state",
            "outputs": [
                {"internalType": "enum ProductState", "name": "", "type": "uint8"}
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "tickSize",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "newOwner", "type": "address"}
            ],
            "name": "transferOwnership",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "newImplementation",
                    "type": "address",
                },
                {"internalType": "bytes", "name": "data", "type": "bytes"},
            ],
            "name": "upgradeToAndCall",
            "outputs": [],
            "stateMutability": "payable",
            "type": "function",
        },
    ],
)



from decimal import Decimal
from functools import cache

from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.exceptions import ContractCustomError

from .. import validators
from ..bindings import (
    ClearingDiamond,
    MarginAccount,
    MarginAccountRegistry,
    ProductRegistry,
)
from ..bindings.erc20 import ERC20
from ..bindings.facade import CLEARING_DIAMOND_ABI
from ..bindings.margin_account import ABI as MARGIN_CONTRACT_ABI
from ..bindings.margin_account_registry import ABI as MARGIN_ACCOUNT_REGISTRY_ABI
from ..bindings.product_registry import ABI as PRODUCT_REGISTRY_ABI
from ..decorators import convert_web3_error
from ..exceptions import NotFoundError
from ..schemas import Position
from .base import ClearingSystemAPI
from .builder import Builder


class Clearing(ClearingSystemAPI):
    """API for managing margin accounts.

    Parameters
    ----------
    private_key : str
        The private key of the blockchain account that manages the margin account.
    autonity_rpc_url : str
        The URL of a JSON-RPC provider for Autonity. (HTTPS only.)
    """

    ### Transactions ###

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def authorize(self, collateral_asset: str, intent_account_id: str) -> str:
        """Authorizes a blockchain account to submit intents to the clearing system
        using the margin account associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        intent_account_id : str
            The address of the intent account.

        Returns
        -------
        str
            The hash of the transaction.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        intent_account_id = validators.validate_address(intent_account_id)

        tx_hash = (
            self._margin_contract(collateral_asset)
            .authorize(intent_account_id)
            .transact()
        )
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def deposit_into_margin_account(
        self, collateral_asset: str, amount: Decimal
    ) -> tuple[str, str]:
        """Deposits the specified amount of collateral tokens into the margin account
        associated with the collateral asset.

        First approves the token transfer with the collateral token, then executes the
        transfer.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        amount : Decimal
            The amount of collateral tokens to deposit.

        Returns
        -------
        str
            The hash of the approval transaction.
        str
            The hash of the deposit transaction.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        token_amount = int(amount * 10 ** self._decimals(collateral_asset))
        token_contract = ERC20(self._w3, collateral_asset)

        tx1_hash = token_contract.approve(
            self._margin_contract(collateral_asset)._contract.address,  # type: ignore
            token_amount,
        ).transact()
        self._w3.eth.wait_for_transaction_receipt(tx1_hash)

        tx2_hash = (
            self._margin_contract(collateral_asset).deposit(token_amount).transact()
        )
        self._w3.eth.wait_for_transaction_receipt(tx2_hash)

        return (Web3.to_hex(tx1_hash), Web3.to_hex(tx2_hash))

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def withdraw_from_margin_account(
        self, collateral_asset: str, amount: Decimal
    ) -> str:
        """Withdraws the specified amount of collateral tokens from the margin account
        associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        amount : Decimal
            The amount of collateral tokens to withdraw.

        Returns
        -------
        str
            The hash of the transaction.
        """
        collateral_asset = validators.validate_address(collateral_asset)
        token_amount = int(amount * 10 ** self._decimals(collateral_asset))
        tx_hash = (
            self._margin_contract(collateral_asset).withdraw(token_amount).transact()
        )
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    @convert_web3_error(CLEARING_DIAMOND_ABI)
    def initiate_final_settlement(self, product_id: str, accounts: list[str]) -> str:
        """Initiate final settlement (closeout) process for the specified accounts.

        The product must be in Final Settlement state. The accounts must hold non-zero
        positions in the product that offset each other (i.e. the sum of their position
        sizes is 0.)

        Parameters
        ----------
        product_id : str
            The ID of the product.
        accounts : list of str
            List of margin account IDs to initiate settlement for.

        Returns
        -------
        str
            The hash of the transaction.
        """
        product_id = validators.validate_hexstr32(product_id)
        addresses = [validators.validate_address(account) for account in accounts]

        clearing_contract = ClearingDiamond(self._w3)
        tx_hash = clearing_contract.initiate_final_settlement(
            HexBytes(product_id), addresses
        ).transact()
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    ### Views ###

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def capital(self, collateral_asset: str) -> Decimal:
        """Returns the amount of collateral tokens in the margin account associated
        with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).capital(self._account.address)
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def position(self, collateral_asset: str, position_id: str) -> Position:
        """Returns the parameters of a position in the margin account associated with
        the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.
        position_id: str
            The ID of the position.

        Returns
        -------
        afp.schemas.Position
        """
        validators.validate_hexstr32(position_id)
        data = self._margin_contract(collateral_asset).position_data(
            self._account.address, HexBytes(position_id)
        )
        decimals = self._decimals(collateral_asset)
        return Position(
            id=Web3.to_hex(data.position_id),
            quantity=data.quantity,
            cost_basis=Decimal(data.cost_basis) / 10**decimals,
            maintenance_margin=Decimal(data.maintenance_margin) / 10**decimals,
            pnl=Decimal(data.pnl) / 10**decimals,
        )

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def positions(self, collateral_asset: str) -> list[Position]:
        """Returns all positions in the margin account associated with the collateral
        asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        list of afp.schemas.Position
        """
        collateral_asset = validators.validate_address(collateral_asset)
        position_ids = self._margin_contract(collateral_asset).positions(
            self._account.address
        )
        return [self.position(collateral_asset, Web3.to_hex(id)) for id in position_ids]

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def margin_account_equity(self, collateral_asset: str) -> Decimal:
        """Returns the margin account equity in the margin account associated with the
        collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mae(self._account.address)
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def maintenance_margin_available(self, collateral_asset: str) -> Decimal:
        """Returns the maintenance margin available in the margin account associated
        with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mma(self._account.address)
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def maintenance_margin_used(self, collateral_asset: str) -> Decimal:
        """Returns the maintenance margin used in the margin account associated with
        the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).mmu(self._account.address)
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def profit_and_loss(self, collateral_asset: str) -> Decimal:
        """Returns the profit and loss in the margin account associated with the
        collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).pnl(self._account.address)
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(MARGIN_CONTRACT_ABI)
    def withdrawable_amount(self, collateral_asset: str) -> Decimal:
        """Returns the amount of collateral tokens withdrawable from the margin account
        associated with the collateral asset.

        Parameters
        ----------
        collateral_asset : str
            The address of the collateral token.

        Returns
        -------
        Decimal
        """
        collateral_asset = validators.validate_address(collateral_asset)
        amount = self._margin_contract(collateral_asset).withdrawable(
            self._account.address
        )
        return Decimal(amount) / 10 ** self._decimals(collateral_asset)

    @convert_web3_error(PRODUCT_REGISTRY_ABI)
    def collateral_asset(self, product_id: str) -> str:
        """Returns the collateral asset of a product.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        str
        """
        product_registry_contract = ProductRegistry(self._w3)
        collateral_asset = product_registry_contract.collateral_asset(
            HexBytes(product_id)
        )
        if Web3.to_int(hexstr=collateral_asset) == 0:
            raise NotFoundError("Product not found in the product registry")
        return collateral_asset

    def product_state(self, product_id: str) -> str:
        """Returns the current state of a product.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        str
        """
        return Builder.product_state(self, product_id)

    ### Internal getters ###

    @cache
    @convert_web3_error(MARGIN_ACCOUNT_REGISTRY_ABI)
    def _margin_contract(self, collateral_asset: ChecksumAddress) -> MarginAccount:
        margin_account_registry_contract = MarginAccountRegistry(self._w3)
        try:
            margin_contract_address = (
                margin_account_registry_contract.get_margin_account(
                    Web3.to_checksum_address(collateral_asset)
                )
            )
        except ContractCustomError:
            raise NotFoundError("No margin account found for collateral asset")
        return MarginAccount(self._w3, margin_contract_address)



"""FinalSettlementFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class FinalSettlementFacet:
    """FinalSettlementFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed FinalSettlementFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    def closeout_fee_rate(
        self,
    ) -> int:
        """Binding for `CLOSEOUT_FEE_RATE` on the FinalSettlementFacet contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.CLOSEOUT_FEE_RATE().call()
        return int(return_value)

    def closeout_reward_rate(
        self,
    ) -> int:
        """Binding for `CLOSEOUT_REWARD_RATE` on the FinalSettlementFacet contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.CLOSEOUT_REWARD_RATE().call()
        return int(return_value)

    def finalize_fsp(
        self,
        product_id: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `finalizeFsp` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.finalizeFsp(
            product_id,
        )

    def get_fsp(
        self,
        product_id: hexbytes.HexBytes,
    ) -> typing.Tuple[int, bool]:
        """Binding for `getFsp` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        bool
        """
        return_value = self._contract.functions.getFsp(
            product_id,
        ).call()
        return (
            int(return_value[0]),
            bool(return_value[1]),
        )

    def initiate_final_settlement(
        self,
        product_id: hexbytes.HexBytes,
        accounts: typing.List[eth_typing.ChecksumAddress],
    ) -> contract.ContractFunction:
        """Binding for `initiateFinalSettlement` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        accounts : typing.List[eth_typing.ChecksumAddress]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initiateFinalSettlement(
            product_id,
            accounts,
        )

    def open_interest(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `openInterest` on the FinalSettlementFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.openInterest(
            product_id,
        ).call()
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "FSPAlreadyFinalized",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "FSPNotFound",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
                {"internalType": "uint256", "name": "currentTime", "type": "uint256"},
                {
                    "internalType": "uint256",
                    "name": "earliestFSPSubmissionTime",
                    "type": "uint256",
                },
            ],
            "name": "FSPTimeNotReached",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "fspAccount", "type": "address"}
            ],
            "name": "InvalidFSPAccount",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "oracleAddress", "type": "address"}
            ],
            "name": "InvalidOracleAddress",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "int256", "name": "checksum", "type": "int256"},
                {
                    "internalType": "int256",
                    "name": "expectedChecksum",
                    "type": "int256",
                },
            ],
            "name": "MismatchedFSPAccountQuantities",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "ProductNotInFinalSettlement",
            "type": "error",
        },
        {
            "inputs": [],
            "name": "CLOSEOUT_FEE_RATE",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "pure",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "CLOSEOUT_REWARD_RATE",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "pure",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productID", "type": "bytes32"}
            ],
            "name": "finalizeFsp",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "getFsp",
            "outputs": [
                {"internalType": "uint256", "name": "fsp", "type": "uint256"},
                {"internalType": "bool", "name": "finalized", "type": "bool"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productID", "type": "bytes32"},
                {"internalType": "address[]", "name": "accounts", "type": "address[]"},
            ],
            "name": "initiateFinalSettlement",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "openInterest",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
    ],
)



"""ClearingFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class Side(enum.IntEnum):
    """Port of `enum Side` on the ClearingFacet contract."""

    BID = 0
    ASK = 1


@dataclass
class AuctionConfig:
    """Port of `struct AuctionConfig` on the IAuctioneer contract."""

    restoration_buffer: int
    liquidation_duration: int


@dataclass
class ClearingConfig:
    """Port of `struct ClearingConfig` on the IClearing contract."""

    clearing_fee_rate: int


@dataclass
class Config:
    """Port of `struct Config` on the IClearing contract."""

    auction_config: AuctionConfig
    clearing_config: ClearingConfig


@dataclass
class IntentData:
    """Port of `struct IntentData` on the IClearing contract."""

    nonce: int
    trading_protocol_id: eth_typing.ChecksumAddress
    product_id: hexbytes.HexBytes
    limit_price: int
    quantity: int
    max_trading_fee_rate: int
    good_until: int
    side: Side


@dataclass
class Intent:
    """Port of `struct Intent` on the IClearing contract."""

    margin_account_id: eth_typing.ChecksumAddress
    intent_account_id: eth_typing.ChecksumAddress
    hash: hexbytes.HexBytes
    data: IntentData
    signature: hexbytes.HexBytes


@dataclass
class Trade:
    """Port of `struct Trade` on the IClearing contract."""

    product_id: hexbytes.HexBytes
    protocol_id: eth_typing.ChecksumAddress
    trade_id: int
    price: int
    timestamp: int
    accounts: typing.List[eth_typing.ChecksumAddress]
    quantities: typing.List[int]
    fee_rates: typing.List[int]
    intents: typing.List[Intent]


class ClearingFacet:
    """ClearingFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ClearingFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def TradeExecuted(self) -> contract.ContractEvent:
        """Binding for `event TradeExecuted` on the ClearingFacet contract."""
        return self._contract.events.TradeExecuted

    def max_trading_fee_rate(
        self,
    ) -> int:
        """Binding for `MAX_TRADING_FEE_RATE` on the ClearingFacet contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.MAX_TRADING_FEE_RATE().call()
        return int(return_value)

    def clearing_fee_rate(
        self,
    ) -> int:
        """Binding for `clearingFeeRate` on the ClearingFacet contract.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.clearingFeeRate().call()
        return int(return_value)

    def config(
        self,
    ) -> Config:
        """Binding for `config` on the ClearingFacet contract.

        Returns
        -------
        Config
        """
        return_value = self._contract.functions.config().call()
        return Config(
            AuctionConfig(int(return_value[0][0]), int(return_value[0][1])),
            ClearingConfig(int(return_value[1][0])),
        )

    def estimate_fees(
        self,
        product_id: hexbytes.HexBytes,
        price: int,
        quantity: int,
        trading_fee_rate: int,
    ) -> typing.Tuple[int, int]:
        """Binding for `estimateFees` on the ClearingFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        price : int
        quantity : int
        trading_fee_rate : int

        Returns
        -------
        int
        int
        """
        return_value = self._contract.functions.estimateFees(
            product_id,
            price,
            quantity,
            trading_fee_rate,
        ).call()
        return (
            int(return_value[0]),
            int(return_value[1]),
        )

    def execute(
        self,
        trade: Trade,
        fallback_on_failure: bool,
    ) -> contract.ContractFunction:
        """Binding for `execute` on the ClearingFacet contract.

        Parameters
        ----------
        trade : Trade
        fallback_on_failure : bool

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.execute(
            (
                trade.product_id,
                trade.protocol_id,
                trade.trade_id,
                trade.price,
                trade.timestamp,
                trade.accounts,
                trade.quantities,
                trade.fee_rates,
                [
                    (
                        trade_item.margin_account_id,
                        trade_item.intent_account_id,
                        trade_item.hash,
                        (
                            trade_item.data.nonce,
                            trade_item.data.trading_protocol_id,
                            trade_item.data.product_id,
                            trade_item.data.limit_price,
                            trade_item.data.quantity,
                            trade_item.data.max_trading_fee_rate,
                            trade_item.data.good_until,
                            int(trade_item.data.side),
                        ),
                        trade_item.signature,
                    )
                    for trade_item in trade.intents
                ],
            ),
            fallback_on_failure,
        )

    def finalize_initialization(
        self,
        margin_account_registry: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `finalizeInitialization` on the ClearingFacet contract.

        Parameters
        ----------
        margin_account_registry : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.finalizeInitialization(
            margin_account_registry,
        )

    def get_admin(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getAdmin` on the ClearingFacet contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getAdmin().call()
        return eth_typing.ChecksumAddress(return_value)

    def get_margin_account_registry(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getMarginAccountRegistry` on the ClearingFacet contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getMarginAccountRegistry().call()
        return eth_typing.ChecksumAddress(return_value)

    def get_product_registry(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getProductRegistry` on the ClearingFacet contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getProductRegistry().call()
        return eth_typing.ChecksumAddress(return_value)

    def get_treasury(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getTreasury` on the ClearingFacet contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getTreasury().call()
        return eth_typing.ChecksumAddress(return_value)

    def hash_intent(
        self,
        intent: Intent,
    ) -> hexbytes.HexBytes:
        """Binding for `hashIntent` on the ClearingFacet contract.

        Parameters
        ----------
        intent : Intent

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.hashIntent(
            (
                intent.margin_account_id,
                intent.intent_account_id,
                intent.hash,
                (
                    intent.data.nonce,
                    intent.data.trading_protocol_id,
                    intent.data.product_id,
                    intent.data.limit_price,
                    intent.data.quantity,
                    intent.data.max_trading_fee_rate,
                    intent.data.good_until,
                    int(intent.data.side),
                ),
                intent.signature,
            ),
        ).call()
        return hexbytes.HexBytes(return_value)

    def initialize(
        self,
        _product_registry: eth_typing.ChecksumAddress,
        _treasury: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the ClearingFacet contract.

        Parameters
        ----------
        _product_registry : eth_typing.ChecksumAddress
        _treasury : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            _product_registry,
            _treasury,
        )

    def is_admin_active(
        self,
    ) -> bool:
        """Binding for `isAdminActive` on the ClearingFacet contract.

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.isAdminActive().call()
        return bool(return_value)

    def set_active(
        self,
        active: bool,
    ) -> contract.ContractFunction:
        """Binding for `setActive` on the ClearingFacet contract.

        Parameters
        ----------
        active : bool

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setActive(
            active,
        )

    def set_admin(
        self,
        new_admin: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `setAdmin` on the ClearingFacet contract.

        Parameters
        ----------
        new_admin : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setAdmin(
            new_admin,
        )

    def set_config(
        self,
        _config: Config,
    ) -> contract.ContractFunction:
        """Binding for `setConfig` on the ClearingFacet contract.

        Parameters
        ----------
        _config : Config

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setConfig(
            (
                (
                    _config.auction_config.restoration_buffer,
                    _config.auction_config.liquidation_duration,
                ),
                (_config.clearing_config.clearing_fee_rate),
            ),
        )

    def version(
        self,
    ) -> str:
        """Binding for `version` on the ClearingFacet contract.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.version().call()
        return str(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "bytes4", "name": "selector", "type": "bytes4"},
                {"internalType": "address", "name": "sender", "type": "address"},
            ],
            "name": "AdminControlledNotAuthorized",
            "type": "error",
        },
        {"inputs": [], "name": "AlreadyInitialized", "type": "error"},
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "DuplicateMarginAccount",
            "type": "error",
        },
        {"inputs": [], "name": "ECDSAInvalidSignature", "type": "error"},
        {
            "inputs": [
                {"internalType": "uint256", "name": "length", "type": "uint256"}
            ],
            "name": "ECDSAInvalidSignatureLength",
            "type": "error",
        },
        {
            "inputs": [{"internalType": "bytes32", "name": "s", "type": "bytes32"}],
            "name": "ECDSAInvalidSignatureS",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "intentAccount", "type": "address"}
            ],
            "name": "IntentFullySpent",
            "type": "error",
        },
        {
            "inputs": [{"internalType": "int256", "name": "feeSum", "type": "int256"}],
            "name": "InvalidFeeSum",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "parameter", "type": "string"}
            ],
            "name": "InvalidIntent",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "enum ProductState", "name": "state", "type": "uint8"}
            ],
            "name": "InvalidProductState",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "signer", "type": "address"},
                {
                    "internalType": "address",
                    "name": "expectedSigner",
                    "type": "address",
                },
            ],
            "name": "InvalidSignature",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "length", "type": "uint256"}
            ],
            "name": "InvalidTradeIntents",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "MAECheckFailed",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "int256", "name": "feeRate", "type": "int256"},
                {"internalType": "uint256", "name": "maxFeeRate", "type": "uint256"},
            ],
            "name": "MaxFeeRateExceeded",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "buySide", "type": "uint256"},
                {"internalType": "uint256", "name": "sellSide", "type": "uint256"},
            ],
            "name": "MismatchedTrade",
            "type": "error",
        },
        {"inputs": [], "name": "QueueIsEmpty", "type": "error"},
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "Unauthorized",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "needed", "type": "address"},
                {"internalType": "address", "name": "got", "type": "address"},
            ],
            "name": "UnauthorizedTradeSubmitter",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "productID",
                    "type": "bytes32",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "protocolID",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "marginAccount",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "price",
                    "type": "uint256",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "quantity",
                    "type": "uint256",
                },
            ],
            "name": "TradeExecuted",
            "type": "event",
        },
        {
            "inputs": [],
            "name": "MAX_TRADING_FEE_RATE",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "pure",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "clearingFeeRate",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "config",
            "outputs": [
                {
                    "components": [
                        {
                            "components": [
                                {
                                    "internalType": "uint64",
                                    "name": "restorationBuffer",
                                    "type": "uint64",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "liquidationDuration",
                                    "type": "uint256",
                                },
                            ],
                            "internalType": "struct IAuctioneer.AuctionConfig",
                            "name": "auctionConfig",
                            "type": "tuple",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "uint32",
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                }
                            ],
                            "internalType": "struct IClearing.ClearingConfig",
                            "name": "clearingConfig",
                            "type": "tuple",
                        },
                    ],
                    "internalType": "struct IClearing.Config",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productID", "type": "bytes32"},
                {"internalType": "uint256", "name": "price", "type": "uint256"},
                {"internalType": "uint256", "name": "quantity", "type": "uint256"},
                {"internalType": "int256", "name": "tradingFeeRate", "type": "int256"},
            ],
            "name": "estimateFees",
            "outputs": [
                {"internalType": "uint256", "name": "clearingFee", "type": "uint256"},
                {"internalType": "int256", "name": "tradingFee", "type": "int256"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "address",
                            "name": "protocolID",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "tradeID",
                            "type": "uint256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "timestamp",
                            "type": "uint256",
                        },
                        {
                            "internalType": "address[]",
                            "name": "accounts",
                            "type": "address[]",
                        },
                        {
                            "internalType": "uint256[]",
                            "name": "quantities",
                            "type": "uint256[]",
                        },
                        {
                            "internalType": "int256[]",
                            "name": "feeRates",
                            "type": "int256[]",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "marginAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "address",
                                    "name": "intentAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "bytes32",
                                    "name": "hash",
                                    "type": "bytes32",
                                },
                                {
                                    "components": [
                                        {
                                            "internalType": "uint256",
                                            "name": "nonce",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "address",
                                            "name": "tradingProtocolID",
                                            "type": "address",
                                        },
                                        {
                                            "internalType": "bytes32",
                                            "name": "productID",
                                            "type": "bytes32",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "limitPrice",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "quantity",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "maxTradingFeeRate",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "goodUntil",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "enum Side",
                                            "name": "side",
                                            "type": "uint8",
                                        },
                                    ],
                                    "internalType": "struct IClearing.IntentData",
                                    "name": "data",
                                    "type": "tuple",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "signature",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IClearing.Intent[]",
                            "name": "intents",
                            "type": "tuple[]",
                        },
                    ],
                    "internalType": "struct IClearing.Trade",
                    "name": "trade",
                    "type": "tuple",
                },
                {"internalType": "bool", "name": "fallbackOnFailure", "type": "bool"},
            ],
            "name": "execute",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountRegistry",
                    "type": "address",
                }
            ],
            "name": "finalizeInitialization",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "getAdmin",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "getMarginAccountRegistry",
            "outputs": [
                {
                    "internalType": "contract IMarginAccountRegistry",
                    "name": "",
                    "type": "address",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "getProductRegistry",
            "outputs": [
                {
                    "internalType": "contract IProductRegistry",
                    "name": "",
                    "type": "address",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "getTreasury",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "internalType": "address",
                            "name": "marginAccountID",
                            "type": "address",
                        },
                        {
                            "internalType": "address",
                            "name": "intentAccountID",
                            "type": "address",
                        },
                        {"internalType": "bytes32", "name": "hash", "type": "bytes32"},
                        {
                            "components": [
                                {
                                    "internalType": "uint256",
                                    "name": "nonce",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "address",
                                    "name": "tradingProtocolID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "bytes32",
                                    "name": "productID",
                                    "type": "bytes32",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "limitPrice",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "quantity",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "maxTradingFeeRate",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "goodUntil",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "enum Side",
                                    "name": "side",
                                    "type": "uint8",
                                },
                            ],
                            "internalType": "struct IClearing.IntentData",
                            "name": "data",
                            "type": "tuple",
                        },
                        {"internalType": "bytes", "name": "signature", "type": "bytes"},
                    ],
                    "internalType": "struct IClearing.Intent",
                    "name": "intent",
                    "type": "tuple",
                }
            ],
            "name": "hashIntent",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "pure",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "_productRegistry",
                    "type": "address",
                },
                {"internalType": "address", "name": "_treasury", "type": "address"},
            ],
            "name": "initialize",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "isAdminActive",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [{"internalType": "bool", "name": "active", "type": "bool"}],
            "name": "setActive",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "newAdmin", "type": "address"}
            ],
            "name": "setAdmin",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "components": [
                                {
                                    "internalType": "uint64",
                                    "name": "restorationBuffer",
                                    "type": "uint64",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "liquidationDuration",
                                    "type": "uint256",
                                },
                            ],
                            "internalType": "struct IAuctioneer.AuctionConfig",
                            "name": "auctionConfig",
                            "type": "tuple",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "uint32",
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                }
                            ],
                            "internalType": "struct IClearing.ClearingConfig",
                            "name": "clearingConfig",
                            "type": "tuple",
                        },
                    ],
                    "internalType": "struct IClearing.Config",
                    "name": "_config",
                    "type": "tuple",
                }
            ],
            "name": "setConfig",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "version",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "stateMutability": "pure",
            "type": "function",
        },
    ],
)



"""BankruptcyFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


@dataclass
class LAAData:
    """Port of `struct LAAData` on the IBankruptcy contract."""

    owner: eth_typing.ChecksumAddress
    quantity: int
    last_traded_timestamp: int
    position_age: int


@dataclass
class PositionLossData:
    """Port of `struct PositionLossData` on the IBankruptcy contract."""

    position_id: hexbytes.HexBytes
    bankrupt_account_u_pn_l: int
    bankrupt_account_quantity: int
    mark_price: int
    point_value: int
    tick_size: int
    laa_data: typing.List[LAAData]


class BankruptcyFacet:
    """BankruptcyFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed BankruptcyFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    def last_traded_timestamp(
        self,
        product_id: hexbytes.HexBytes,
        trader: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `lastTradedTimestamp` on the BankruptcyFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        trader : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.lastTradedTimestamp(
            product_id,
            trader,
        ).call()
        return int(return_value)

    def mutualize_losses(
        self,
        bankrupt_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
        product_ids: typing.List[hexbytes.HexBytes],
        laas: typing.List[typing.List[eth_typing.ChecksumAddress]],
    ) -> contract.ContractFunction:
        """Binding for `mutualizeLosses` on the BankruptcyFacet contract.

        Parameters
        ----------
        bankrupt_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress
        product_ids : typing.List[hexbytes.HexBytes]
        laas : typing.List[typing.List[eth_typing.ChecksumAddress]]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.mutualizeLosses(
            bankrupt_account_id,
            collateral_token,
            product_ids,
            laas,
        )

    def unique_trader_count(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `uniqueTraderCount` on the BankruptcyFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.uniqueTraderCount(
            product_id,
        ).call()
        return int(return_value)

    def validate_la_as(
        self,
        margin_account: eth_typing.ChecksumAddress,
        bankrupt_account: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        product_owners: typing.List[eth_typing.ChecksumAddress],
    ) -> PositionLossData:
        """Binding for `validateLAAs` on the BankruptcyFacet contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        bankrupt_account : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        product_owners : typing.List[eth_typing.ChecksumAddress]

        Returns
        -------
        PositionLossData
        """
        return_value = self._contract.functions.validateLAAs(
            margin_account,
            bankrupt_account,
            product_id,
            product_owners,
        ).call()
        return PositionLossData(
            hexbytes.HexBytes(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            int(return_value[4]),
            int(return_value[5]),
            [
                LAAData(
                    eth_typing.ChecksumAddress(return_value6_elem[0]),
                    int(return_value6_elem[1]),
                    int(return_value6_elem[2]),
                    int(return_value6_elem[3]),
                )
                for return_value6_elem in return_value[6]
            ],
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "AccountNotBankrupt",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "int256", "name": "neededQuantity", "type": "int256"},
                {
                    "internalType": "int256",
                    "name": "totalLAAQuantity",
                    "type": "int256",
                },
            ],
            "name": "LAAQuantityMismatch",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "neededCount", "type": "uint256"},
                {
                    "internalType": "uint256",
                    "name": "totalLAASetCount",
                    "type": "uint256",
                },
            ],
            "name": "LAASetIncomplete",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "NoTradeData",
            "type": "error",
        },
        {"inputs": [], "name": "QueueIsEmpty", "type": "error"},
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
                {"internalType": "address", "name": "trader", "type": "address"},
            ],
            "name": "lastTradedTimestamp",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "bankruptAccountId",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
                {
                    "internalType": "bytes32[]",
                    "name": "productIds",
                    "type": "bytes32[]",
                },
                {"internalType": "address[][]", "name": "laas", "type": "address[][]"},
            ],
            "name": "mutualizeLosses",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "uniqueTraderCount",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "contract IMarginAccount",
                    "name": "marginAccount",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "bankruptAccount",
                    "type": "address",
                },
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
                {
                    "internalType": "address[]",
                    "name": "productOwners",
                    "type": "address[]",
                },
            ],
            "name": "validateLAAs",
            "outputs": [
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "bankruptAccountUPnL",
                            "type": "int256",
                        },
                        {
                            "internalType": "int256",
                            "name": "bankruptAccountQuantity",
                            "type": "int256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "markPrice",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "pointValue",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "tickSize",
                            "type": "uint256",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "owner",
                                    "type": "address",
                                },
                                {
                                    "internalType": "int256",
                                    "name": "quantity",
                                    "type": "int256",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "lastTradedTimestamp",
                                    "type": "uint256",
                                },
                                {
                                    "internalType": "uint256",
                                    "name": "positionAge",
                                    "type": "uint256",
                                },
                            ],
                            "internalType": "struct IBankruptcy.LAAData[]",
                            "name": "laaData",
                            "type": "tuple[]",
                        },
                    ],
                    "internalType": "struct IBankruptcy.PositionLossData",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
    ],
)



from datetime import datetime, UTC

import pytest

from afp import validators
from afp.schemas import OrderFillFilter


def test_timestamp_conversion():
    dt_time = datetime.fromisoformat("2030-01-01T12:00:00Z")
    ts_time = 1893499200

    # timestamp -> datetime
    filter = OrderFillFilter(
        intent_account_id="",
        product_id=None,
        margin_account_id=None,
        intent_hash=None,
        start=None,
        end=ts_time,  # type: ignore
        trade_state=None,
    )
    assert filter.end is not None
    assert filter.end.astimezone(UTC) == dt_time

    # datetime -> timestamp
    assert '"end":%d' % ts_time in filter.model_dump_json(exclude_none=True)


def test_validate_hexstr32__pass():
    assert validators.validate_hexstr32(
        "0xe50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b"
    )


@pytest.mark.parametrize(
    "value",
    [
        "e50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b",
        "0xg50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b",
        "0xe50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb6",
    ],
    ids=str,
)
def test_validate_hexstr32__error(value):
    with pytest.raises(ValueError):
        validators.validate_hexstr32(value)



"""Typed bindings around the smart contracts of the Autonomous Futures Protocol."""

from .auctioneer_facet import AuctionConfig, AuctionData, BidData
from .bankruptcy_facet import LAAData, PositionLossData
from .clearing_facet import ClearingConfig, Config, Intent, IntentData, Side, Trade
from .facade import (
    ClearingDiamond,
    MarginAccountRegistry,
    OracleProvider,
    ProductRegistry,
)
from .margin_account import MarginAccount, PositionData, Settlement
from .product_registry import (
    OracleSpecification,
    Product,
    ProductMetadata,
    ProductState,
)
from .trading_protocol import TradingProtocol

__all__ = (
    # Contract bindings
    "ClearingDiamond",
    "MarginAccount",
    "MarginAccountRegistry",
    "OracleProvider",
    "ProductRegistry",
    "TradingProtocol",
    # Structures
    "AuctionConfig",
    "AuctionData",
    "BidData",
    "ClearingConfig",
    "Config",
    "Intent",
    "IntentData",
    "LAAData",
    "OracleSpecification",
    "PositionData",
    "PositionLossData",
    "Product",
    "ProductMetadata",
    "Settlement",
    "Trade",
    # Enumerations
    "ProductState",
    "Side",
)



"""MarkPriceTrackerFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class MarkPriceTrackerFacet:
    """MarkPriceTrackerFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarkPriceTrackerFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    def valuation(
        self,
        product_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `valuation` on the MarkPriceTrackerFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.valuation(
            product_id,
        ).call()
        return int(return_value)

    def valuation_after_trade(
        self,
        product_id: hexbytes.HexBytes,
        price: int,
        quantity: int,
    ) -> int:
        """Binding for `valuationAfterTrade` on the MarkPriceTrackerFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        price : int
        quantity : int

        Returns
        -------
        int
        """
        return_value = self._contract.functions.valuationAfterTrade(
            product_id,
            price,
            quantity,
        ).call()
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "NoTradeData",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "valuation",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
                {"internalType": "uint256", "name": "price", "type": "uint256"},
                {"internalType": "uint256", "name": "quantity", "type": "uint256"},
            ],
            "name": "valuationAfterTrade",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
    ],
)



from datetime import datetime, timedelta
from decimal import Decimal
from pprint import pprint
from time import sleep

import afp


MARGIN_ACCOUNT_ID = "0x79EFD85867d4Ae3a96a65d66707266647d771023"
INTENT_ACCOUNT_PRIVATE_KEY = (
    "0xdbacc0d8d0b5dc20a7a68f9ceb2daa3d5dc7ab43b06d2eda4b9a41a08be60024"
)

PRODUCT_ID = "0xf82118deb932a8649d519d8d34e7f7b278a44bdb3f2663f6049aaea6ee33b211"


def main():
    # Trader submits bid
    trading = afp.Trading(INTENT_ACCOUNT_PRIVATE_KEY)

    intent = trading.create_intent(
        margin_account_id=MARGIN_ACCOUNT_ID,
        product=trading.product(PRODUCT_ID),
        side="bid",
        limit_price=Decimal("2.34"),
        quantity=1,
        max_trading_fee_rate=Decimal("0.1"),
        good_until_time=datetime.now() + timedelta(hours=1),
    )
    limit_order = trading.submit_limit_order(intent)
    pprint(limit_order.model_dump())

    # Trader cancels the bid
    cancel_order = trading.submit_cancel_order(intent.hash)
    pprint(cancel_order.model_dump())

    # Check that the order is cancelled
    sleep(0.5)
    pprint(trading.order(limit_order.id).model_dump())


if __name__ == "__main__":
    main()
